=== modified file 'Makefile'
--- Makefile	2011-10-15 16:48:03 +0000
+++ Makefile	2011-11-24 21:12:35 +0000
@@ -250,7 +250,7 @@
 	@echo "###################################################################"
 	./plugin-runner --plugin-dir=plugins.d \
 		--config-file=plugin-runner.conf \
-		--options-for=mandos-client:--seckey=keydir/seckey.txt,--pubkey=keydir/pubkey.txt \
+		--options-for=mandos-client:--seckey=keydir/seckey.txt,--pubkey=keydir/pubkey.txt,--network-hook-dir=network-hooks.d \
 		$(CLIENTARGS)
 
 # Used by run-client

=== modified file 'debian/mandos-client.README.Debian'
--- debian/mandos-client.README.Debian	2011-10-05 16:00:56 +0000
+++ debian/mandos-client.README.Debian	2011-11-27 14:44:28 +0000
@@ -51,6 +51,11 @@
   disk environment, the network interface *must* exist at that stage.
   Thus, the interface can *not* be a pseudo-interface such as "br0" or
   "tun0"; instead, only real interface (such as "eth0") can be used.
+  This can be overcome by writing a "network hook" program to create
+  the interface (see mandos-client(8mandos)) and placing it in
+  "/etc/mandos/network-hooks.d", from where it will be copied into the
+  initial RAM disk.  Example network hook scripts can be found in
+  "/usr/share/doc/mandos-client/network-hooks.d".
 
 * User-Supplied Plugins
   
@@ -84,4 +89,4 @@
   work, "--options-for=mandos-client:--connect=<ADDRESS>:<PORT>" needs
   to be manually added to the file "/etc/mandos/plugin-runner.conf".
 
- -- Teddy Hogeborn <teddy@recompile.se>, Wed,  5 Oct 2011 17:50:22 +0200
+ -- Teddy Hogeborn <teddy@fukt.bsnet.se>, Sun, 27 Nov 2011 15:06:34 +0100

=== modified file 'debian/mandos-client.docs'
--- debian/mandos-client.docs	2008-10-18 11:17:22 +0000
+++ debian/mandos-client.docs	2011-11-27 02:32:20 +0000
@@ -1,3 +1,4 @@
 NEWS
 README
 TODO
+network-hooks.d

=== modified file 'debian/rules'
--- debian/rules	2010-09-09 18:16:14 +0000
+++ debian/rules	2011-11-27 14:44:28 +0000
@@ -85,7 +85,8 @@
 	dh_fixperms --exclude etc/keys/mandos \
 		--exclude etc/mandos/clients.conf \
 		--exclude etc/mandos/plugins.d \
-		--exclude usr/lib/mandos/plugins.d
+		--exclude usr/lib/mandos/plugins.d \
+		--exclude /usr/share/doc/mandos-client/network-hooks.d
 	dh_installdeb
 	dh_shlibdeps
 	dh_gencontrol

=== modified file 'initramfs-tools-hook'
--- initramfs-tools-hook	2011-11-13 20:49:21 +0000
+++ initramfs-tools-hook	2011-11-27 14:44:28 +0000
@@ -72,7 +72,7 @@
 
 # Make directories
 install --directory --mode=u=rwx,go=rx "${DESTDIR}${CONFDIR}" \
-	"${DESTDIR}${MANDOSDIR}"
+	"${DESTDIR}${MANDOSDIR}" "${DESTDIR}${HOOKDIR}"
 install --owner=${mandos_user} --group=${mandos_group} --directory \
     --mode=u=rwx "${DESTDIR}${PLUGINDIR}"
 
@@ -111,18 +111,27 @@
 for hook in /etc/mandos/network-hooks.d/*; do
     case "`basename \"$hook\"`" in
 	"*") continue ;;
-	*[!A-Za-z0-9_-]*) continue ;;
+	*[!A-Za-z0-9_.-]*) continue ;;
 	*) test -d "$hook" || copy_exec "$hook" "${HOOKDIR}" ;;
     esac
-    # Copy any files needed by the network hook
-    MANDOSNETHOOKDIR=/etc/mandos/network-hooks.d MODE=files \
-	"$hook" files | while read file target; do
-	if [ -z "${target}" ]; then
-	    copy_exec "$file"
-	else
-	    copy_exec "$file" "$target"
-	fi
-    done
+    if [ -x "$hook" ]; then
+        # Copy any files needed by the network hook
+	MANDOSNETHOOKDIR=/etc/mandos/network-hooks.d MODE=files \
+	    VERBOSITY=0 "$hook" files | while read file target; do
+	    if [ -z "${target}" ]; then
+		copy_exec "$file"
+	    else
+		copy_exec "$file" "$target"
+	    fi
+	done
+        # Copy and load any modules needed by the network hook
+	MANDOSNETHOOKDIR=/etc/mandos/network-hooks.d MODE=modules \
+	    VERBOSITY=0 "$hook" modules | while read module; do
+	    if [ -z "${target}" ]; then
+		force_load "$module"
+	    fi
+	done
+    fi
 done
 
 # GPGME needs /usr/bin/gpg

=== added directory 'network-hooks.d'
=== added file 'network-hooks.d/bridge'
--- network-hooks.d/bridge	1970-01-01 00:00:00 +0000
+++ network-hooks.d/bridge	2011-11-27 14:44:28 +0000
@@ -0,0 +1,64 @@
+#!/bin/sh
+#
+# This is an example of a Mandos client network hook.  This hook
+# brings up a bridge interface as specified in a separate
+# configuration file.  To be used, this file and any needed
+# configuration file(s) should be copied into the
+# /etc/mandos/network-hooks.d directory.
+# 
+# Copying and distribution of this file, with or without modification,
+# are permitted in any medium without royalty provided the copyright
+# notice and this notice are preserved.  This file is offered as-is,
+# without any warranty.
+
+set -e
+
+CONFIG="$MANDOSNETHOOKDIR/bridge.conf"
+
+# Read config file, which must set "BRIDGE", "PORTS", and optionally
+# "IPADDRS" and "ROUTES".
+if [ -e "$CONFIG" ]; then
+    . "$CONFIG"
+fi
+
+if [ -z "$BRIDGE" -o -z "$PORTS" ]; then
+    exit
+fi
+
+if [ -n "$DEVICE" -a "$DEVICE" != "$BRIDGE" ]; then
+    exit
+fi
+
+case "$1" in
+    start)
+	brctl addbr "$BRIDGE"
+	for port in $PORTS; do
+	    brctl addif "$BRIDGE" "$port"
+	done
+	ip link set up "$BRIDGE"
+	if [ -n "$IPADDRS" ]; then
+            for ipaddr in $IPADDRS; do
+		ip addr add "$ipaddr" dev "$BRIDGE"
+	    done
+	fi
+	if [ -n "$ROUTES" ]; then
+            for route in $ROUTES; do
+		ip route add "$route" dev "$BRIDGE"
+	    done
+	fi
+	;;
+    stop)
+	ip link set down "$BRIDGE"
+	for port in $PORTS; do
+	    brctl delif "$BRIDGE" "$port"
+	done
+	brctl delbr "$BRIDGE"
+	;;
+    files)
+	echo /bin/ip
+	echo /usr/sbin/brctl
+	;;
+    modules)
+	echo bridge
+	;;
+esac

=== added file 'network-hooks.d/bridge.conf'
--- network-hooks.d/bridge.conf	1970-01-01 00:00:00 +0000
+++ network-hooks.d/bridge.conf	2011-11-24 20:15:24 +0000
@@ -0,0 +1,11 @@
+## Required
+
+#BRIDGE=br0
+
+#PORTS="eth0 eth1"
+
+## Optional
+
+#IPADDRS="192.0.2.3/24 2001:DB8::aede:48ff:fe71:f6f2/32"
+
+#ROUTES="192.0.2.0/24 2001:DB8::/32"

=== modified file 'plugin-runner.c'
--- plugin-runner.c	2011-10-09 12:32:13 +0000
+++ plugin-runner.c	2011-11-24 21:12:35 +0000
@@ -742,7 +742,7 @@
     }
   }
   
-  {
+  if(getuid() == 0){
     /* Work around Debian bug #633582:
        <http://bugs.debian.org/633582> */
     int plugindir_fd = open(/* plugindir or */ PDIR, O_RDONLY);

=== modified file 'plugins.d/mandos-client.c'
--- plugins.d/mandos-client.c	2011-11-14 18:32:54 +0000
+++ plugins.d/mandos-client.c	2011-11-24 21:29:56 +0000
@@ -178,7 +178,8 @@
   va_list ap;
   va_start (ap, format);
   
-  TEMP_FAILURE_RETRY(fprintf(stream, "Mandos plugin %s: ", program_invocation_short_name));
+  TEMP_FAILURE_RETRY(fprintf(stream, "Mandos plugin %s: ",
+			     program_invocation_short_name));
   return TEMP_FAILURE_RETRY(vfprintf(stream, format, ap));
 }
 
@@ -261,17 +262,15 @@
     
     rc = gpgme_data_new_from_fd(&pgp_data, fd);
     if(rc != GPG_ERR_NO_ERROR){
-      fprintf(stderr, "Mandos plugin mandos-client: "
-	      "bad gpgme_data_new_from_fd: %s: %s\n",
-	      gpgme_strsource(rc), gpgme_strerror(rc));
+      fprintf_plus(stderr, "bad gpgme_data_new_from_fd: %s: %s\n",
+		   gpgme_strsource(rc), gpgme_strerror(rc));
       return false;
     }
     
     rc = gpgme_op_import(mc.ctx, pgp_data);
     if(rc != GPG_ERR_NO_ERROR){
-      fprintf(stderr, "Mandos plugin mandos-client: "
-	      "bad gpgme_op_import: %s: %s\n",
-	      gpgme_strsource(rc), gpgme_strerror(rc));
+      fprintf_plus(stderr, "bad gpgme_op_import: %s: %s\n",
+		   gpgme_strsource(rc), gpgme_strerror(rc));
       return false;
     }
     
@@ -284,26 +283,23 @@
   }
   
   if(debug){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "Initializing GPGME\n");
+    fprintf_plus(stderr, "Initializing GPGME\n");
   }
   
   /* Init GPGME */
   gpgme_check_version(NULL);
   rc = gpgme_engine_check_version(GPGME_PROTOCOL_OpenPGP);
   if(rc != GPG_ERR_NO_ERROR){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "bad gpgme_engine_check_version: %s: %s\n",
-	    gpgme_strsource(rc), gpgme_strerror(rc));
+    fprintf_plus(stderr, "bad gpgme_engine_check_version: %s: %s\n",
+		 gpgme_strsource(rc), gpgme_strerror(rc));
     return false;
   }
   
   /* Set GPGME home directory for the OpenPGP engine only */
   rc = gpgme_get_engine_info(&engine_info);
   if(rc != GPG_ERR_NO_ERROR){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "bad gpgme_get_engine_info: %s: %s\n",
-	    gpgme_strsource(rc), gpgme_strerror(rc));
+    fprintf_plus(stderr, "bad gpgme_get_engine_info: %s: %s\n",
+		 gpgme_strsource(rc), gpgme_strerror(rc));
     return false;
   }
   while(engine_info != NULL){
@@ -315,17 +311,17 @@
     engine_info = engine_info->next;
   }
   if(engine_info == NULL){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "Could not set GPGME home dir to %s\n", tempdir);
+    fprintf_plus(stderr, "Could not set GPGME home dir to %s\n",
+		 tempdir);
     return false;
   }
   
   /* Create new GPGME "context" */
   rc = gpgme_new(&(mc.ctx));
   if(rc != GPG_ERR_NO_ERROR){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "bad gpgme_new: %s: %s\n", gpgme_strsource(rc),
-	    gpgme_strerror(rc));
+    fprintf_plus(stderr, "Mandos plugin mandos-client: "
+		 "bad gpgme_new: %s: %s\n", gpgme_strsource(rc),
+		 gpgme_strerror(rc));
     return false;
   }
   
@@ -350,26 +346,24 @@
   ssize_t plaintext_length = 0;
   
   if(debug){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "Trying to decrypt OpenPGP data\n");
+    fprintf_plus(stderr, "Trying to decrypt OpenPGP data\n");
   }
   
   /* Create new GPGME data buffer from memory cryptotext */
   rc = gpgme_data_new_from_mem(&dh_crypto, cryptotext, crypto_size,
 			       0);
   if(rc != GPG_ERR_NO_ERROR){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "bad gpgme_data_new_from_mem: %s: %s\n",
-	    gpgme_strsource(rc), gpgme_strerror(rc));
+    fprintf_plus(stderr, "bad gpgme_data_new_from_mem: %s: %s\n",
+		 gpgme_strsource(rc), gpgme_strerror(rc));
     return -1;
   }
   
   /* Create new empty GPGME data buffer for the plaintext */
   rc = gpgme_data_new(&dh_plain);
   if(rc != GPG_ERR_NO_ERROR){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "bad gpgme_data_new: %s: %s\n",
-	    gpgme_strsource(rc), gpgme_strerror(rc));
+    fprintf_plus(stderr, "Mandos plugin mandos-client: "
+		 "bad gpgme_data_new: %s: %s\n",
+		 gpgme_strsource(rc), gpgme_strerror(rc));
     gpgme_data_release(dh_crypto);
     return -1;
   }
@@ -378,39 +372,32 @@
      data buffer */
   rc = gpgme_op_decrypt(mc.ctx, dh_crypto, dh_plain);
   if(rc != GPG_ERR_NO_ERROR){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "bad gpgme_op_decrypt: %s: %s\n",
-	    gpgme_strsource(rc), gpgme_strerror(rc));
+    fprintf_plus(stderr, "bad gpgme_op_decrypt: %s: %s\n",
+		 gpgme_strsource(rc), gpgme_strerror(rc));
     plaintext_length = -1;
     if(debug){
       gpgme_decrypt_result_t result;
       result = gpgme_op_decrypt_result(mc.ctx);
       if(result == NULL){
-	fprintf(stderr, "Mandos plugin mandos-client: "
-		"gpgme_op_decrypt_result failed\n");
+	fprintf_plus(stderr, "gpgme_op_decrypt_result failed\n");
       } else {
-	fprintf(stderr, "Mandos plugin mandos-client: "
-		"Unsupported algorithm: %s\n",
-		result->unsupported_algorithm);
-	fprintf(stderr, "Mandos plugin mandos-client: "
-		"Wrong key usage: %u\n",
-		result->wrong_key_usage);
+	fprintf_plus(stderr, "Unsupported algorithm: %s\n",
+		     result->unsupported_algorithm);
+	fprintf_plus(stderr, "Wrong key usage: %u\n",
+		     result->wrong_key_usage);
 	if(result->file_name != NULL){
-	  fprintf(stderr, "Mandos plugin mandos-client: "
-		  "File name: %s\n", result->file_name);
+	  fprintf_plus(stderr, "File name: %s\n", result->file_name);
 	}
 	gpgme_recipient_t recipient;
 	recipient = result->recipients;
 	while(recipient != NULL){
-	  fprintf(stderr, "Mandos plugin mandos-client: "
-		  "Public key algorithm: %s\n",
-		  gpgme_pubkey_algo_name(recipient->pubkey_algo));
-	  fprintf(stderr, "Mandos plugin mandos-client: "
-		  "Key ID: %s\n", recipient->keyid);
-	  fprintf(stderr, "Mandos plugin mandos-client: "
-		  "Secret key available: %s\n",
-		  recipient->status == GPG_ERR_NO_SECKEY
-		  ? "No" : "Yes");
+	  fprintf_plus(stderr, "Public key algorithm: %s\n",
+		       gpgme_pubkey_algo_name
+		       (recipient->pubkey_algo));
+	  fprintf_plus(stderr, "Key ID: %s\n", recipient->keyid);
+	  fprintf_plus(stderr, "Secret key available: %s\n",
+		       recipient->status == GPG_ERR_NO_SECKEY
+		       ? "No" : "Yes");
 	  recipient = recipient->next;
 	}
       }
@@ -419,8 +406,7 @@
   }
   
   if(debug){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "Decryption of OpenPGP data succeeded\n");
+    fprintf_plus(stderr, "Decryption of OpenPGP data succeeded\n");
   }
   
   /* Seek back to the beginning of the GPGME plaintext data buffer */
@@ -457,8 +443,7 @@
   }
   
   if(debug){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "Decrypted password is: ");
+    fprintf_plus(stderr, "Decrypted password is: ");
     for(ssize_t i = 0; i < plaintext_length; i++){
       fprintf(stderr, "%02hhX ", (*plaintext)[i]);
     }
@@ -486,7 +471,7 @@
 /* GnuTLS log function callback */
 static void debuggnutls(__attribute__((unused)) int level,
 			const char* string){
-  fprintf(stderr, "Mandos plugin mandos-client: GnuTLS: %s", string);
+  fprintf_plus(stderr, "GnuTLS: %s", string);
 }
 
 static int init_gnutls_global(const char *pubkeyfilename,
@@ -494,14 +479,13 @@
   int ret;
   
   if(debug){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "Initializing GnuTLS\n");
+    fprintf_plus(stderr, "Initializing GnuTLS\n");
   }
   
   ret = gnutls_global_init();
   if(ret != GNUTLS_E_SUCCESS){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "GnuTLS global_init: %s\n", safer_gnutls_strerror(ret));
+    fprintf_plus(stderr, "GnuTLS global_init: %s\n",
+		 safer_gnutls_strerror(ret));
     return -1;
   }
   
@@ -516,45 +500,43 @@
   /* OpenPGP credentials */
   ret = gnutls_certificate_allocate_credentials(&mc.cred);
   if(ret != GNUTLS_E_SUCCESS){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "GnuTLS memory error: %s\n", safer_gnutls_strerror(ret));
+    fprintf_plus(stderr, "GnuTLS memory error: %s\n",
+		 safer_gnutls_strerror(ret));
     gnutls_global_deinit();
     return -1;
   }
   
   if(debug){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "Attempting to use OpenPGP public key %s and"
-	    " secret key %s as GnuTLS credentials\n", pubkeyfilename,
-	    seckeyfilename);
+    fprintf_plus(stderr, "Attempting to use OpenPGP public key %s and"
+		 " secret key %s as GnuTLS credentials\n",
+		 pubkeyfilename,
+		 seckeyfilename);
   }
   
   ret = gnutls_certificate_set_openpgp_key_file
     (mc.cred, pubkeyfilename, seckeyfilename,
      GNUTLS_OPENPGP_FMT_BASE64);
   if(ret != GNUTLS_E_SUCCESS){
-    fprintf(stderr,
-	    "Mandos plugin mandos-client: "
-	    "Error[%d] while reading the OpenPGP key pair ('%s',"
-	    " '%s')\n", ret, pubkeyfilename, seckeyfilename);
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "The GnuTLS error is: %s\n", safer_gnutls_strerror(ret));
+    fprintf_plus(stderr,
+		 "Error[%d] while reading the OpenPGP key pair ('%s',"
+		 " '%s')\n", ret, pubkeyfilename, seckeyfilename);
+    fprintf_plus(stderr, "The GnuTLS error is: %s\n",
+		 safer_gnutls_strerror(ret));
     goto globalfail;
   }
   
   /* GnuTLS server initialization */
   ret = gnutls_dh_params_init(&mc.dh_params);
   if(ret != GNUTLS_E_SUCCESS){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "Error in GnuTLS DH parameter initialization:"
-	    " %s\n", safer_gnutls_strerror(ret));
+    fprintf_plus(stderr, "Error in GnuTLS DH parameter"
+		 " initialization: %s\n",
+		 safer_gnutls_strerror(ret));
     goto globalfail;
   }
   ret = gnutls_dh_params_generate2(mc.dh_params, mc.dh_bits);
   if(ret != GNUTLS_E_SUCCESS){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "Error in GnuTLS prime generation: %s\n",
-	    safer_gnutls_strerror(ret));
+    fprintf_plus(stderr, "Error in GnuTLS prime generation: %s\n",
+		 safer_gnutls_strerror(ret));
     goto globalfail;
   }
   
@@ -580,9 +562,9 @@
     }
   } while(ret == GNUTLS_E_INTERRUPTED or ret == GNUTLS_E_AGAIN);
   if(ret != GNUTLS_E_SUCCESS){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "Error in GnuTLS session initialization: %s\n",
-	    safer_gnutls_strerror(ret));
+    fprintf_plus(stderr,
+		 "Error in GnuTLS session initialization: %s\n",
+		 safer_gnutls_strerror(ret));
   }
   
   {
@@ -595,10 +577,9 @@
       }
     } while(ret == GNUTLS_E_INTERRUPTED or ret == GNUTLS_E_AGAIN);
     if(ret != GNUTLS_E_SUCCESS){
-      fprintf(stderr, "Mandos plugin mandos-client: "
-	      "Syntax error at: %s\n", err);
-      fprintf(stderr, "Mandos plugin mandos-client: "
-	      "GnuTLS error: %s\n", safer_gnutls_strerror(ret));
+      fprintf_plus(stderr, "Syntax error at: %s\n", err);
+      fprintf_plus(stderr, "GnuTLS error: %s\n",
+		   safer_gnutls_strerror(ret));
       gnutls_deinit(*session);
       return -1;
     }
@@ -613,9 +594,8 @@
     }
   } while(ret == GNUTLS_E_INTERRUPTED or ret == GNUTLS_E_AGAIN);
   if(ret != GNUTLS_E_SUCCESS){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "Error setting GnuTLS credentials: %s\n",
-	    safer_gnutls_strerror(ret));
+    fprintf_plus(stderr, "Error setting GnuTLS credentials: %s\n",
+		 safer_gnutls_strerror(ret));
     gnutls_deinit(*session);
     return -1;
   }
@@ -666,8 +646,7 @@
     pf = PF_INET;
     break;
   default:
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "Bad address family: %d\n", af);
+    fprintf_plus(stderr, "Bad address family: %d\n", af);
     errno = EINVAL;
     return -1;
   }
@@ -678,9 +657,8 @@
   }
   
   if(debug){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "Setting up a TCP connection to %s, port %" PRIu16
-	    "\n", ip, port);
+    fprintf_plus(stderr, "Setting up a TCP connection to %s, port %"
+		 PRIu16 "\n", ip, port);
   }
   
   tcp_sd = socket(pf, SOCK_STREAM, 0);
@@ -712,8 +690,7 @@
   }
   if(ret == 0){
     int e = errno;
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "Bad address: %s\n", ip);
+    fprintf_plus(stderr, "Bad address: %s\n", ip);
     errno = e;
     goto mandos_end;
   }
@@ -726,9 +703,8 @@
        (&to.in6.sin6_addr)){ /* -Wstrict-aliasing=2 or lower and
 				-Wunreachable-code*/
       if(if_index == AVAHI_IF_UNSPEC){
-	fprintf(stderr, "Mandos plugin mandos-client: "
-		"An IPv6 link-local address is incomplete"
-		" without a network interface\n");
+	fprintf_plus(stderr, "An IPv6 link-local address is"
+		     " incomplete without a network interface\n");
 	errno = EINVAL;
 	goto mandos_end;
       }
@@ -752,13 +728,12 @@
       if(if_indextoname((unsigned int)if_index, interface) == NULL){
 	perror_plus("if_indextoname");
       } else {
-	fprintf(stderr, "Mandos plugin mandos-client: "
-		"Connection to: %s%%%s, port %" PRIu16 "\n",
-		ip, interface, port);
+	fprintf_plus(stderr, "Connection to: %s%%%s, port %" PRIu16
+		     "\n", ip, interface, port);
       }
     } else {
-      fprintf(stderr, "Mandos plugin mandos-client: "
-	      "Connection to: %s, port %" PRIu16 "\n", ip, port);
+      fprintf_plus(stderr, "Connection to: %s, port %" PRIu16 "\n",
+		   ip, port);
     }
     char addrstr[(INET_ADDRSTRLEN > INET6_ADDRSTRLEN) ?
 		 INET_ADDRSTRLEN : INET6_ADDRSTRLEN] = "";
@@ -774,8 +749,7 @@
       perror_plus("inet_ntop");
     } else {
       if(strcmp(addrstr, ip) != 0){
-	fprintf(stderr, "Mandos plugin mandos-client: "
-		"Canonical address form: %s\n", addrstr);
+	fprintf_plus(stderr, "Canonical address form: %s\n", addrstr);
       }
     }
   }
@@ -835,8 +809,7 @@
   }
   
   if(debug){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "Establishing TLS session with %s\n", ip);
+    fprintf_plus(stderr, "Establishing TLS session with %s\n", ip);
   }
   
   if(quit_now){
@@ -862,8 +835,7 @@
   
   if(ret != GNUTLS_E_SUCCESS){
     if(debug){
-      fprintf(stderr, "Mandos plugin mandos-client: "
-	      "*** GnuTLS Handshake failed ***\n");
+      fprintf_plus(stderr, "*** GnuTLS Handshake failed ***\n");
       gnutls_perror(ret);
     }
     errno = EPROTO;
@@ -873,8 +845,8 @@
   /* Read OpenPGP packet that contains the wanted password */
   
   if(debug){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "Retrieving OpenPGP encrypted password from %s\n", ip);
+    fprintf_plus(stderr, "Retrieving OpenPGP encrypted password from"
+		 " %s\n", ip);
   }
   
   while(true){
@@ -918,17 +890,16 @@
 	  }
 	} while(ret == GNUTLS_E_AGAIN or ret == GNUTLS_E_INTERRUPTED);
 	if(ret < 0){
-	  fprintf(stderr, "Mandos plugin mandos-client: "
-		  "*** GnuTLS Re-handshake failed ***\n");
+	  fprintf_plus(stderr, "*** GnuTLS Re-handshake failed "
+		       "***\n");
 	  gnutls_perror(ret);
 	  errno = EPROTO;
 	  goto mandos_end;
 	}
 	break;
       default:
-	fprintf(stderr, "Mandos plugin mandos-client: "
-		"Unknown error while reading data from"
-		" encrypted session with Mandos server\n");
+	fprintf_plus(stderr, "Unknown error while reading data from"
+		     " encrypted session with Mandos server\n");
 	gnutls_bye(session, GNUTLS_SHUT_RDWR);
 	errno = EIO;
 	goto mandos_end;
@@ -939,8 +910,7 @@
   }
   
   if(debug){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "Closing TLS session\n");
+    fprintf_plus(stderr, "Closing TLS session\n");
   }
   
   if(quit_now){
@@ -975,9 +945,8 @@
 	if(ret == 0 and ferror(stdout)){
 	  int e = errno;
 	  if(debug){
-	    fprintf(stderr, "Mandos plugin mandos-client: "
-		    "Error writing encrypted data: %s\n",
-		    strerror(errno));
+	    fprintf_plus(stderr, "Error writing encrypted data: %s\n",
+			 strerror(errno));
 	  }
 	  errno = e;
 	  goto mandos_end;
@@ -1040,10 +1009,10 @@
   switch(event){
   default:
   case AVAHI_RESOLVER_FAILURE:
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "(Avahi Resolver) Failed to resolve service '%s'"
-	    " of type '%s' in domain '%s': %s\n", name, type, domain,
-	    avahi_strerror(avahi_server_errno(mc.server)));
+    fprintf_plus(stderr, "(Avahi Resolver) Failed to resolve service "
+		 "'%s' of type '%s' in domain '%s': %s\n", name, type,
+		 domain,
+		 avahi_strerror(avahi_server_errno(mc.server)));
     break;
     
   case AVAHI_RESOLVER_FOUND:
@@ -1051,10 +1020,9 @@
       char ip[AVAHI_ADDRESS_STR_MAX];
       avahi_address_snprint(ip, sizeof(ip), address);
       if(debug){
-	fprintf(stderr, "Mandos plugin mandos-client: "
-		"Mandos server \"%s\" found on %s (%s, %"
-		PRIdMAX ") on port %" PRIu16 "\n", name, host_name,
-		ip, (intmax_t)interface, port);
+	fprintf_plus(stderr, "Mandos server \"%s\" found on %s (%s, %"
+		     PRIdMAX ") on port %" PRIu16 "\n", name,
+		     host_name, ip, (intmax_t)interface, port);
       }
       int ret = start_mandos_communication(ip, port, interface,
 					   avahi_proto_to_af(proto));
@@ -1092,9 +1060,8 @@
   default:
   case AVAHI_BROWSER_FAILURE:
     
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "(Avahi browser) %s\n",
-	    avahi_strerror(avahi_server_errno(mc.server)));
+    fprintf_plus(stderr, "(Avahi browser) %s\n",
+		 avahi_strerror(avahi_server_errno(mc.server)));
     avahi_simple_poll_quit(mc.simple_poll);
     return;
     
@@ -1107,9 +1074,9 @@
     if(avahi_s_service_resolver_new(mc.server, interface, protocol,
 				    name, type, domain, protocol, 0,
 				    resolve_callback, NULL) == NULL)
-      fprintf(stderr, "Mandos plugin mandos-client: "
-	      "Avahi: Failed to resolve service '%s': %s\n",
-	      name, avahi_strerror(avahi_server_errno(mc.server)));
+      fprintf_plus(stderr, "Avahi: Failed to resolve service '%s':"
+		   " %s\n", name,
+		   avahi_strerror(avahi_server_errno(mc.server)));
     break;
     
   case AVAHI_BROWSER_REMOVE:
@@ -1118,8 +1085,8 @@
   case AVAHI_BROWSER_ALL_FOR_NOW:
   case AVAHI_BROWSER_CACHE_EXHAUSTED:
     if(debug){
-      fprintf(stderr, "Mandos plugin mandos-client: "
-	      "No Mandos server found, still searching...\n");
+      fprintf_plus(stderr, "No Mandos server found, still"
+		   " searching...\n");
     }
     break;
   }
@@ -1164,40 +1131,39 @@
   /* Reject the loopback device */
   if(ifr->ifr_flags & IFF_LOOPBACK){
     if(debug){
-      fprintf(stderr, "Mandos plugin mandos-client: "
-	      "Rejecting loopback interface \"%s\"\n", ifname);
+      fprintf_plus(stderr, "Rejecting loopback interface \"%s\"\n",
+		   ifname);
     }
     return false;
   }
   /* Accept point-to-point devices only if connect_to is specified */
   if(connect_to != NULL and (ifr->ifr_flags & IFF_POINTOPOINT)){
     if(debug){
-      fprintf(stderr, "Mandos plugin mandos-client: "
-	      "Accepting point-to-point interface \"%s\"\n", ifname);
+      fprintf_plus(stderr, "Accepting point-to-point interface"
+		   " \"%s\"\n", ifname);
     }
     return true;
   }
   /* Otherwise, reject non-broadcast-capable devices */
   if(not (ifr->ifr_flags & IFF_BROADCAST)){
     if(debug){
-      fprintf(stderr, "Mandos plugin mandos-client: "
-	      "Rejecting non-broadcast interface \"%s\"\n", ifname);
+      fprintf_plus(stderr, "Rejecting non-broadcast interface"
+		   " \"%s\"\n", ifname);
     }
     return false;
   }
   /* Reject non-ARP interfaces (including dummy interfaces) */
   if(ifr->ifr_flags & IFF_NOARP){
     if(debug){
-      fprintf(stderr, "Mandos plugin mandos-client: "
-	      "Rejecting non-ARP interface \"%s\"\n", ifname);
+      fprintf_plus(stderr, "Rejecting non-ARP interface \"%s\"\n",
+		   ifname);
     }
     return false;
   }
   
   /* Accept this device */
   if(debug){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "Interface \"%s\" is good\n", ifname);
+    fprintf_plus(stderr, "Interface \"%s\" is good\n", ifname);
   }
   return true;
 }
@@ -1215,9 +1181,8 @@
   struct ifreq ifr;
   if(not get_flags(if_entry->d_name, &ifr)){
     if(debug){
-      fprintf(stderr, "Mandos plugin mandos-client: "
-	      "Failed to get flags for interface \"%s\"\n",
-	      if_entry->d_name);
+      fprintf_plus(stderr, "Failed to get flags for interface "
+		   "\"%s\"\n", if_entry->d_name);
     }
     return 0;
   }
@@ -1241,9 +1206,8 @@
   struct ifreq ifr;
   if(not get_flags(if_entry->d_name, &ifr)){
     if(debug){
-      fprintf(stderr, "Mandos plugin mandos-client: "
-	      "Failed to get flags for interface \"%s\"\n",
-	      if_entry->d_name);
+      fprintf_plus(stderr, "Failed to get flags for interface "
+		   "\"%s\"\n", if_entry->d_name);
     }
     return 0;
   }
@@ -1251,9 +1215,8 @@
   /* Reject down interfaces */
   if(not (ifr.ifr_flags & IFF_UP)){
     if(debug){
-      fprintf(stderr, "Mandos plugin mandos-client: "
-	      "Rejecting down interface \"%s\"\n",
-	      if_entry->d_name);
+      fprintf_plus(stderr, "Rejecting down interface \"%s\"\n",
+		   if_entry->d_name);
     }
     return 0;
   }
@@ -1261,9 +1224,8 @@
   /* Reject non-running interfaces */
   if(not (ifr.ifr_flags & IFF_RUNNING)){
     if(debug){
-      fprintf(stderr, "Mandos plugin mandos-client: "
-	      "Rejecting non-running interface \"%s\"\n",
-	      if_entry->d_name);
+      fprintf_plus(stderr, "Rejecting non-running interface \"%s\"\n",
+		   if_entry->d_name);
     }
     return 0;
   }
@@ -1303,9 +1265,8 @@
   if((direntry->d_name)[sret] != '\0'){
     /* Contains non-allowed characters */
     if(debug){
-      fprintf(stderr, "Mandos plugin mandos-client: "
-	      "Ignoring hook \"%s\" with bad name\n",
-	      direntry->d_name);
+      fprintf_plus(stderr, "Ignoring hook \"%s\" with bad name\n",
+		   direntry->d_name);
     }
     return 0;
   }
@@ -1327,21 +1288,23 @@
   if(not (S_ISREG(st.st_mode))){
     /* Not a regular file */
     if(debug){
-      fprintf(stderr, "Mandos plugin mandos-client: "
-	      "Ignoring hook \"%s\" - not a file\n",
-	      direntry->d_name);
+      fprintf_plus(stderr, "Ignoring hook \"%s\" - not a file\n",
+		   direntry->d_name);
     }
     return 0;
   }
   if(not (st.st_mode & (S_IXUSR | S_IXGRP | S_IXOTH))){
     /* Not executable */
     if(debug){
-      fprintf(stderr, "Mandos plugin mandos-client: "
-	      "Ignoring hook \"%s\" - not executable\n",
-	      direntry->d_name);
+      fprintf_plus(stderr, "Ignoring hook \"%s\" - not executable\n",
+		   direntry->d_name);
     }
     return 0;
   }
+  if(debug){
+    fprintf_plus(stderr, "Hook \"%s\" is acceptable\n",
+		 direntry->d_name);
+  }
   return 1;
 }
 
@@ -1354,15 +1317,14 @@
   while(true){
     if(mc.current_server == NULL){
       if (debug){
-	fprintf(stderr, "Mandos plugin mandos-client: "
-		"Wait until first server is found. No timeout!\n");
+	fprintf_plus(stderr, "Wait until first server is found."
+		     " No timeout!\n");
       }
       ret = avahi_simple_poll_iterate(s, -1);
     } else {
       if (debug){
-	fprintf(stderr, "Mandos plugin mandos-client: "
-		"Check current_server if we should run it,"
-		" or wait\n");
+	fprintf_plus(stderr, "Check current_server if we should run"
+		     " it, or wait\n");
       }
       /* the current time */
       ret = clock_gettime(CLOCK_MONOTONIC, &now);
@@ -1384,8 +1346,8 @@
 		    - ((intmax_t)waited_time.tv_nsec / 1000000));
       
       if (debug){
-	fprintf(stderr, "Mandos plugin mandos-client: "
-		"Blocking for %" PRIdMAX " ms\n", block_time);
+	fprintf_plus(stderr, "Blocking for %" PRIdMAX " ms\n",
+		     block_time);
       }
       
       if(block_time <= 0){
@@ -1418,6 +1380,109 @@
   }
 }
 
+bool run_network_hooks(const char *mode, const char *interface,
+		       const float delay){
+  struct dirent **direntries;
+  struct dirent *direntry;
+  int ret;
+  int numhooks = scandir(hookdir, &direntries, runnable_hook,
+			 alphasort);
+  if(numhooks == -1){
+    perror_plus("scandir");
+  } else {
+    int devnull = open("/dev/null", O_RDONLY);
+    for(int i = 0; i < numhooks; i++){
+      direntry = direntries[i];
+      char *fullname = NULL;
+      ret = asprintf(&fullname, "%s/%s", hookdir, direntry->d_name);
+      if(ret < 0){
+	perror_plus("asprintf");
+	continue;
+      }
+      if(debug){
+	fprintf_plus(stderr, "Running network hook \"%s\"\n",
+		     direntry->d_name);
+      }
+      pid_t hook_pid = fork();
+      if(hook_pid == 0){
+	/* Child */
+	dup2(devnull, STDIN_FILENO);
+	close(devnull);
+	dup2(STDERR_FILENO, STDOUT_FILENO);
+	ret = setenv("MANDOSNETHOOKDIR", hookdir, 1);
+	if(ret == -1){
+	  perror_plus("setenv");
+	  return false;
+	}
+	ret = setenv("DEVICE", interface, 1);
+	if(ret == -1){
+	  perror_plus("setenv");
+	  return false;
+	}
+	ret = setenv("VERBOSE", debug ? "1" : "0", 1);
+	if(ret == -1){
+	  perror_plus("setenv");
+	  return false;
+	}
+	ret = setenv("MODE", mode, 1);
+	if(ret == -1){
+	  perror_plus("setenv");
+	  return false;
+	}
+	char *delaystring;
+	ret = asprintf(&delaystring, "%f", delay);
+	if(ret == -1){
+	  perror_plus("asprintf");
+	  return false;
+	}
+	ret = setenv("DELAY", delaystring, 1);
+	if(ret == -1){
+	  free(delaystring);
+	  perror_plus("setenv");
+	  return false;
+	}
+	free(delaystring);
+	ret = execl(fullname, direntry->d_name, mode, NULL);
+	perror_plus("execl");
+      } else {
+	int status;
+	if(TEMP_FAILURE_RETRY(waitpid(hook_pid, &status, 0)) == -1){
+	  perror_plus("waitpid");
+	  free(fullname);
+	  continue;
+	}
+	if(WIFEXITED(status)){
+	  if(WEXITSTATUS(status) != 0){
+	    fprintf_plus(stderr, "Warning: network hook \"%s\" exited"
+			 " with status %d\n", direntry->d_name,
+			 WEXITSTATUS(status));
+	    free(fullname);
+	    continue;
+	  }
+	} else if(WIFSIGNALED(status)){
+	  fprintf_plus(stderr, "Warning: network hook \"%s\" died by"
+		       " signal %d\n", direntry->d_name,
+		       WTERMSIG(status));
+	  free(fullname);
+	  continue;
+	} else {
+	  fprintf_plus(stderr, "Warning: network hook \"%s\""
+		       " crashed\n", direntry->d_name);
+	  free(fullname);
+	  continue;
+	}
+      }
+      free(fullname);
+      if(debug){
+	fprintf_plus(stderr, "Network hook \"%s\" ran successfully\n",
+		     direntry->d_name);
+      }
+    }
+    close(devnull);
+  }
+  return true;
+}
+
 int main(int argc, char *argv[]){
   AvahiSServiceBrowser *sb = NULL;
   int error;
@@ -1581,8 +1646,9 @@
 	argp_state_help(state, state->out_stream,
 			ARGP_HELP_USAGE | ARGP_HELP_EXIT_ERR);
       case 'V':			/* --version */
-	fprintf(state->out_stream, "Mandos plugin mandos-client: ");
-	fprintf(state->out_stream, "%s\n", argp_program_version);
+	fprintf_plus(state->out_stream,
+		     "Mandos plugin mandos-client: ");
+	fprintf_plus(state->out_stream, "%s\n", argp_program_version);
 	exit(argp_err_exit_status);
 	break;
       default:
@@ -1612,7 +1678,7 @@
     }
   }
     
-  {
+  if(getuid() == 0){
     /* Work around Debian bug #633582:
        <http://bugs.debian.org/633582> */
     struct stat st;
@@ -1633,7 +1699,8 @@
 	if(ret == -1){
 	  perror_plus("fstat");
 	} else {
-	  if(S_ISREG(st.st_mode) and st.st_uid == 0 and st.st_gid == 0){
+	  if(S_ISREG(st.st_mode)
+	     and st.st_uid == 0 and st.st_gid == 0){
 	    ret = fchown(seckey_fd, uid, gid);
 	    if(ret == -1){
 	      perror_plus("fchown");
@@ -1653,7 +1720,8 @@
 	if(ret == -1){
 	  perror_plus("fstat");
 	} else {
-	  if(S_ISREG(st.st_mode) and st.st_uid == 0 and st.st_gid == 0){
+	  if(S_ISREG(st.st_mode)
+	     and st.st_uid == 0 and st.st_gid == 0){
 	    ret = fchown(pubkey_fd, uid, gid);
 	    if(ret == -1){
 	      perror_plus("fchown");
@@ -1672,102 +1740,26 @@
     }
   }
   
-  /* Find network hooks and run them */
+  /* Run network hooks */
   {
-    struct dirent **direntries;
-    struct dirent *direntry;
-    int numhooks = scandir(hookdir, &direntries, runnable_hook,
-			   alphasort);
-    if(numhooks == -1){
-      perror_plus("scandir");
-    } else {
-      int devnull = open("/dev/null", O_RDONLY);
-      for(int i = 0; i < numhooks; i++){
-	direntry = direntries[0];
-	char *fullname = NULL;
-	ret = asprintf(&fullname, "%s/%s", hookdir, direntry->d_name);
-	if(ret < 0){
-	  perror_plus("asprintf");
-	  continue;
-	}
-	pid_t hook_pid = fork();
-	if(hook_pid == 0){
-	  /* Child */
-	  dup2(devnull, STDIN_FILENO);
-	  close(devnull);
-	  dup2(STDERR_FILENO, STDOUT_FILENO);
-	  ret = setenv("MANDOSNETHOOKDIR", hookdir, 1);
-	  if(ret == -1){
-	    perror_plus("setenv");
-	    exit(1);
-	  }
-	  ret = setenv("DEVICE", interface, 1);
-	  if(ret == -1){
-	    perror_plus("setenv");
-	    exit(1);
-	  }
-	  ret = setenv("VERBOSE", debug ? "1" : "0", 1);
-	  if(ret == -1){
-	    perror_plus("setenv");
-	    exit(1);
-	  }
-	  ret = setenv("MODE", "start", 1);
-	  if(ret == -1){
-	    perror_plus("setenv");
-	    exit(1);
-	  }
-	  char *delaystring;
-	  ret = asprintf(&delaystring, "%f", delay);
-	  if(ret == -1){
-	    perror_plus("asprintf");
-	    exit(1);
-	  }
-	  ret = setenv("DELAY", delaystring, 1);
-	  if(ret == -1){
-	    free(delaystring);
-	    perror_plus("setenv");
-	    exit(1);
-	  }
-	  free(delaystring);
-	  ret = execl(fullname, direntry->d_name, "start", NULL);
-	  perror_plus("execl");
-	} else {
-	  int status;
-	  if(TEMP_FAILURE_RETRY(waitpid(hook_pid, &status, 0)) == -1){
-	    perror_plus("waitpid");
-	    free(fullname);
-	    continue;
-	  }
-	  if(WIFEXITED(status)){
-	    if(WEXITSTATUS(status) != 0){
-	      fprintf(stderr, "Mandos plugin mandos-client: "
-		      "Warning: network hook \"%s\" exited"
-		      " with status %d\n", direntry->d_name,
-		      WEXITSTATUS(status));
-	      free(fullname);
-	      continue;
-	    }
-	  } else if(WIFSIGNALED(status)){
-	    fprintf(stderr, "Mandos plugin mandos-client: "
-		    "Warning: network hook \"%s\" died by"
-		    " signal %d\n", direntry->d_name,
-		    WTERMSIG(status));
-	    free(fullname);
-	    continue;
-	  } else {
-	    fprintf(stderr, "Mandos plugin mandos-client: "
-		    "Warning: network hook \"%s\" crashed\n",
-		    direntry->d_name);
-	    free(fullname);
-	    continue;
-	  }
-	}
-	free(fullname);
-	if(quit_now){
-	  goto end;
-	}
-      }
-      close(devnull);
+    if(getuid() == 0){
+      /* Re-raise priviliges */
+      errno = 0;
+      ret = seteuid(0);
+      if(ret == -1){
+	perror_plus("seteuid");
+      }
+    }
+    if(not run_network_hooks("start", interface, delay)){
+      goto end;
+    }
+    if(getuid() == 0){
+      /* Lower privileges */
+      errno = 0;
+      ret = seteuid(uid);
+      if(ret == -1){
+	perror_plus("seteuid");
+      }
     }
   }
   
@@ -1790,8 +1782,7 @@
       /* Pick the first interface returned */
       interface = strdup(direntries[0]->d_name);
       if(debug){
-	fprintf(stderr, "Mandos plugin mandos-client: "
-		"Using interface \"%s\"\n", interface);
+	fprintf_plus(stderr, "Using interface \"%s\"\n", interface);
       }
       if(interface == NULL){
 	perror_plus("malloc");
@@ -1802,8 +1793,7 @@
       free(direntries);
     } else {
       free(direntries);
-      fprintf(stderr, "Mandos plugin mandos-client: "
-	      "Could not find a network interface\n");
+      fprintf_plus(stderr, "Could not find a network interface\n");
       exitcode = EXIT_FAILURE;
       goto end;
     }
@@ -1815,8 +1805,8 @@
   srand((unsigned int) time(NULL));
   mc.simple_poll = avahi_simple_poll_new();
   if(mc.simple_poll == NULL){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "Avahi: Failed to create simple poll object.\n");
+    fprintf_plus(stderr,
+		 "Avahi: Failed to create simple poll object.\n");
     exitcode = EX_UNAVAILABLE;
     goto end;
   }
@@ -1888,8 +1878,7 @@
   if(strcmp(interface, "none") != 0){
     if_index = (AvahiIfIndex) if_nametoindex(interface);
     if(if_index == 0){
-      fprintf(stderr, "Mandos plugin mandos-client: "
-	      "No such interface: \"%s\"\n", interface);
+      fprintf_plus(stderr, "No such interface: \"%s\"\n", interface);
       exitcode = EX_UNAVAILABLE;
       goto end;
     }
@@ -2015,18 +2004,10 @@
 #endif	/* __linux__ */
     /* Lower privileges */
     errno = 0;
-    if(take_down_interface){
-      /* Lower privileges */
-      ret = seteuid(uid);
-      if(ret == -1){
-	perror_plus("seteuid");
-      }
-    } else {
-      /* Lower privileges permanently */
-      ret = setuid(uid);
-      if(ret == -1){
-	perror_plus("setuid");
-      }
+    /* Lower privileges */
+    ret = seteuid(uid);
+    if(ret == -1){
+      perror_plus("seteuid");
     }
   }
   
@@ -2036,8 +2017,7 @@
   
   ret = init_gnutls_global(pubkey, seckey);
   if(ret == -1){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "init_gnutls_global failed\n");
+    fprintf_plus(stderr, "init_gnutls_global failed\n");
     exitcode = EX_UNAVAILABLE;
     goto end;
   } else {
@@ -2059,8 +2039,7 @@
   }
   
   if(not init_gpgme(pubkey, seckey, tempdir)){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "init_gpgme failed\n");
+    fprintf_plus(stderr, "init_gpgme failed\n");
     exitcode = EX_UNAVAILABLE;
     goto end;
   } else {
@@ -2076,8 +2055,7 @@
     /* (Mainly meant for debugging) */
     char *address = strrchr(connect_to, ':');
     if(address == NULL){
-      fprintf(stderr, "Mandos plugin mandos-client: "
-	      "No colon in address\n");
+      fprintf_plus(stderr, "No colon in address\n");
       exitcode = EX_USAGE;
       goto end;
     }
@@ -2091,8 +2069,7 @@
     tmpmax = strtoimax(address+1, &tmp, 10);
     if(errno != 0 or tmp == address+1 or *tmp != '\0'
        or tmpmax != (uint16_t)tmpmax){
-      fprintf(stderr, "Mandos plugin mandos-client: "
-	      "Bad port number\n");
+      fprintf_plus(stderr, "Bad port number\n");
       exitcode = EX_USAGE;
       goto end;
     }
@@ -2128,8 +2105,8 @@
 	break;
       }
       if(debug){
-	fprintf(stderr, "Mandos plugin mandos-client: "
-		"Retrying in %d seconds\n", (int)retry_interval);
+	fprintf_plus(stderr, "Retrying in %d seconds\n",
+		     (int)retry_interval);
       }
       sleep((int)retry_interval);
     }
@@ -2165,9 +2142,8 @@
   
   /* Check if creating the Avahi server object succeeded */
   if(mc.server == NULL){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "Failed to create Avahi server: %s\n",
-	    avahi_strerror(error));
+    fprintf_plus(stderr, "Failed to create Avahi server: %s\n",
+		 avahi_strerror(error));
     exitcode = EX_UNAVAILABLE;
     goto end;
   }
@@ -2181,9 +2157,8 @@
 				   AVAHI_PROTO_UNSPEC, "_mandos._tcp",
 				   NULL, 0, browse_callback, NULL);
   if(sb == NULL){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "Failed to create service browser: %s\n",
-	    avahi_strerror(avahi_server_errno(mc.server)));
+    fprintf_plus(stderr, "Failed to create service browser: %s\n",
+		 avahi_strerror(avahi_server_errno(mc.server)));
     exitcode = EX_UNAVAILABLE;
     goto end;
   }
@@ -2195,23 +2170,20 @@
   /* Run the main loop */
   
   if(debug){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "Starting Avahi loop search\n");
+    fprintf_plus(stderr, "Starting Avahi loop search\n");
   }
 
   ret = avahi_loop_with_timeout(mc.simple_poll,
 				(int)(retry_interval * 1000));
   if(debug){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "avahi_loop_with_timeout exited %s\n",
-	    (ret == 0) ? "successfully" : "with error");
+    fprintf_plus(stderr, "avahi_loop_with_timeout exited %s\n",
+		 (ret == 0) ? "successfully" : "with error");
   }
   
  end:
   
   if(debug){
-    fprintf(stderr, "Mandos plugin mandos-client: "
-	    "%s exiting\n", argv[0]);
+    fprintf_plus(stderr, "%s exiting\n", argv[0]);
   }
   
   /* Cleanup things */
@@ -2233,7 +2205,7 @@
   if(gpgme_initialized){
     gpgme_release(mc.ctx);
   }
-
+  
   /* Cleans up the circular linked list of Mandos servers the client
      has seen */
   if(mc.current_server != NULL){
@@ -2245,17 +2217,21 @@
     }
   }
   
-  /* XXX run network hooks "stop" here  */
-  
-  /* Take down the network interface */
-  if(take_down_interface){
-    /* Re-raise priviliges */
-    errno = 0;
-    ret = seteuid(0);
-    if(ret == -1){
-      perror_plus("seteuid");
+  /* Re-raise priviliges */
+  {
+    if(getuid() == 0){
+      errno = 0;
+      ret = seteuid(0);
+      if(ret == -1){
+	perror_plus("seteuid");
+      }
     }
-    if(geteuid() == 0){
+    
+    /* Run network hooks */
+    run_network_hooks("stop", interface, delay);
+    
+    /* Take down the network interface */
+    if(take_down_interface and geteuid() == 0){
       ret = ioctl(sd, SIOCGIFFLAGS, &network);
       if(ret == -1){
 	perror_plus("ioctl SIOCGIFFLAGS");
@@ -2270,12 +2246,14 @@
       if(ret == -1){
 	perror_plus("close");
       }
-      /* Lower privileges permanently */
-      errno = 0;
-      ret = setuid(uid);
-      if(ret == -1){
-	perror_plus("setuid");
-      }
+    }
+  }
+  if(getuid() == 0){
+    /* Lower privileges permanently */
+    errno = 0;
+    ret = setuid(uid);
+    if(ret == -1){
+      perror_plus("setuid");
     }
   }
   
@@ -2297,8 +2275,8 @@
 	}
 	ret = remove(fullname);
 	if(ret == -1){
-	  fprintf(stderr, "Mandos plugin mandos-client: "
-		  "remove(\"%s\"): %s\n", fullname, strerror(errno));
+	  fprintf_plus(stderr, "remove(\"%s\"): %s\n", fullname,
+		       strerror(errno));
 	}
 	free(fullname);
       }

=== modified file 'plugins.d/mandos-client.xml'
--- plugins.d/mandos-client.xml	2011-11-13 20:49:21 +0000
+++ plugins.d/mandos-client.xml	2011-11-27 14:44:28 +0000
@@ -2,7 +2,7 @@
 <!DOCTYPE refentry PUBLIC "-//OASIS//DTD DocBook XML V4.5//EN"
 	"http://www.oasis-open.org/docbook/xml/4.5/docbookx.dtd" [
 <!ENTITY COMMANDNAME "mandos-client">
-<!ENTITY TIMESTAMP "2011-10-03">
+<!ENTITY TIMESTAMP "2011-11-27">
 <!ENTITY % common SYSTEM "../common.ent">
 %common;
 ]>
@@ -102,7 +102,8 @@
       </arg>
       <sbr/>
       <arg>
-	<option>--network-hook-dir<replaceable>DIR</replaceable></option>
+	<option>--network-hook-dir
+	<replaceable>DIR</replaceable></option>
       </arg>
       <sbr/>
       <arg>
@@ -147,6 +148,26 @@
       will wait indefinitely for new servers to appear.
     </para>
     <para>
+      The network interface is selected like this: If an interface is
+      specified using the <option>--interface</option> option, that
+      interface is used.  Otherwise, <command>&COMMANDNAME;</command>
+      will choose any interface that is up and running and is not a
+      loopback interface, is not a point-to-point interface, is
+      capable of broadcasting and does not have the NOARP flag (see
+      <citerefentry><refentrytitle>netdevice</refentrytitle>
+      <manvolnum>7</manvolnum></citerefentry>).  (If the
+      <option>--connect</option> option is used, point-to-point
+      interfaces and non-broadcast interfaces are accepted.)  If no
+      acceptable interfaces are found, re-run the check but without
+      the <quote>up and running</quote> requirement, and manually take
+      the selected interface up (and later take it down on program
+      exit).
+    </para>
+    <para>
+      Before a network interface is selected, all <quote>network
+      hooks</quote> are run; see <xref linkend="network-hooks"/>.
+    </para>
+    <para>
       This program is not meant to be run directly; it is really meant
       to run as a plugin of the <application>Mandos</application>
       <citerefentry><refentrytitle>plugin-runner</refentrytitle>
@@ -227,7 +248,8 @@
 	    can not be a pseudo-interface such as <quote>br0</quote>
 	    or <quote>tun0</quote>; such interfaces will not exist
 	    until much later in the boot process, and can not be used
-	    by this program.
+	    by this program, unless created by a <quote>network
+	    hook</quote> — see <xref linkend="network-hooks"/>.
 	  </para>
 	  <para>
 	    <replaceable>NAME</replaceable> can be the string
@@ -393,8 +415,10 @@
       <refentrytitle>plugin-runner</refentrytitle>
       <manvolnum>8mandos</manvolnum></citerefentry>) is used to run
       both this program and others in in parallel,
-      <emphasis>one</emphasis> of which will prompt for passwords on
-      the system console.
+      <emphasis>one</emphasis> of which (<citerefentry>
+      <refentrytitle>password-prompt</refentrytitle>
+      <manvolnum>8mandos</manvolnum></citerefentry>) will prompt for
+      passwords on the system console.
     </para>
   </refsect1>
   
@@ -421,6 +445,158 @@
     </para>
   </refsect1>
   
+  <refsect1 id="network-hooks">
+    <title>NETWORK HOOKS</title>
+    <para>
+      If a network interface like a bridge or tunnel is required to
+      find a Mandos server, this requires the interface to be up and
+      running before <command>&COMMANDNAME;</command> starts looking
+      for Mandos servers.  This can be accomplished by creating a
+      <quote>network hook</quote> program, and placing it in a special
+      directory.
+    </para>
+    <para>
+      Before the network is used (and again before program exit), any
+      runnable programs found in the network hook directory are run
+      with the argument <quote><literal>start</literal></quote> or
+      <quote><literal>stop</literal></quote>.  This should bring up or
+      down, respectively, any network interface which
+      <command>&COMMANDNAME;</command> should use.
+    </para>
+    <refsect2 id="hook-requirements">
+      <title>REQUIREMENTS</title>
+      <para>
+	A network hook must be an executable file, and its name must
+	consist entirely of upper and lower case letters, digits,
+	underscores, periods, and hyphens.
+      </para>
+      <para>
+	A network hook will receive one argument, which can be one of
+	the following:
+      </para>
+      <variablelist>
+	<varlistentry>
+	  <term><literal>start</literal></term>
+	  <listitem>
+	    <para>
+	      This should make the network hook create (if necessary)
+	      and bring up a network interface.
+	    </para>
+	  </listitem>
+	</varlistentry>
+	<varlistentry>
+	  <term><literal>stop</literal></term>
+	  <listitem>
+	    <para>
+	      This should make the network hook take down a network
+	      interface, and delete it if it did not exist previously.
+	    </para>
+	  </listitem>
+	</varlistentry>
+	<varlistentry>
+	  <term><literal>files</literal></term>
+	  <listitem>
+	    <para>
+	      This should make the network hook print, <emphasis>one
+	      file per line</emphasis>, all the files needed for it to
+	      run.  (These files will be copied into the initial RAM
+	      filesystem.)  Typical use is for a network hook which is
+	      a shell script to print its needed binaries.
+	    </para>
+	    <para>
+	      It is not necessary to print any non-executable files
+	      already in the network hook directory, these will be
+	      copied implicitly if they otherwise satisfy the name
+	      requirement.
+	    </para>
+	  </listitem>
+	</varlistentry>
+	<varlistentry>
+	  <term><literal>modules</literal></term>
+	  <listitem>
+	    <para>
+	      This should make the network hook print, <emphasis>on
+	      separate lines</emphasis>, all the kernel modules needed
+	      for it to run.  (These modules will be copied into the
+	      initial RAM filesystem.)  For instance, a tunnel
+	      interface needs the
+	      <quote><literal>tun</literal></quote> module.
+	    </para>
+	  </listitem>
+	</varlistentry>
+      </variablelist>
+      <para>
+	The network hook will be provided with a number of environment
+	variables:
+      </para>
+      <variablelist>
+	<varlistentry>
+	  <term><envar>MANDOSNETHOOKDIR</envar></term>
+	  <listitem>
+	    <para>
+	      The network hook directory, specified to
+	      <command>&COMMANDNAME;</command> by the
+	      <option>--network-hook-dir</option> option.  Note: this
+	      should <emphasis>always</emphasis> be used by the
+	      network hook to refer to itself or any files in the hook
+	      directory it may require.
+	    </para>
+	  </listitem>
+	</varlistentry>
+	<varlistentry>
+	  <term><envar>DEVICE</envar></term>
+	  <listitem>
+	    <para>
+	      The network interface, as specified to
+	      <command>&COMMANDNAME;</command> by the
+	      <option>--interface</option> option.  If this is not the
+	      interface a hook will bring up, there is no reason for a
+	      hook to continue.
+	    </para>
+	  </listitem>
+	</varlistentry>
+	<varlistentry>
+	  <term><envar>MODE</envar></term>
+	  <listitem>
+	    <para>
+	      This will be the same as the first argument;
+	      i.e. <quote><literal>start</literal></quote>,
+	      <quote><literal>stop</literal></quote>,
+	      <quote><literal>files</literal></quote>, or
+	      <quote><literal>modules</literal></quote>.
+	    </para>
+	  </listitem>
+	</varlistentry>
+	<varlistentry>
+	  <term><envar>VERBOSITY</envar></term>
+	  <listitem>
+	    <para>
+	      This will be the <quote><literal>1</literal></quote> if
+	      the <option>--debug</option> option is passed to
+	      <command>&COMMANDNAME;</command>, otherwise
+	      <quote><literal>0</literal></quote>.
+	    </para>
+	  </listitem>
+	</varlistentry>
+	<varlistentry>
+	  <term><envar>DELAY</envar></term>
+	  <listitem>
+	    <para>
+	      This will be the same as the <option>--delay</option>
+	      option passed to <command>&COMMANDNAME;</command>.
+	    </para>
+	  </listitem>
+	</varlistentry>
+      </variablelist>
+      <para>
+	A hook may not read from standard input, and should be
+	restrictive in printing to standard output or standard error
+	unless <varname>VERBOSITY</varname> is
+	<quote><literal>1</literal></quote>.
+      </para>
+    </refsect2>
+  </refsect1>
+  
   <refsect1 id="files">
     <title>FILES</title>
     <variablelist>
@@ -438,6 +614,17 @@
 	  </para>
 	</listitem>
       </varlistentry>
+      <varlistentry>
+	<term><filename
+	class="directory">/lib/mandos/network-hooks.d</filename></term>
+	<listitem>
+	  <para>
+	    Directory where network hooks are located.  Change this
+	    with the <option>--network-hook-dir</option> option.  See
+	    <xref linkend="network-hooks"/>.
+	  </para>
+	</listitem>
+      </varlistentry>
     </variablelist>
   </refsect1>
   

